/**********************************************************************************************
 * Filename:       RSC_Service.c
 *
 * Description:    This file contains the implementation of the service.
 *
 * Copyright (c) 2015-2019, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *************************************************************************************************/


/*********************************************************************
 * INCLUDES
 */
#include <string.h>

#include <icall.h>
/* This Header file contains all BLE API and icall structure definition */
#include "icall_ble_api.h"

#include "RSC_Service.h"




/*********************************************************************
 * MACROS
 */

/*********************************************************************
 * CONSTANTS
 */

/*********************************************************************
 * TYPEDEFS
 */

/*********************************************************************
* GLOBAL VARIABLES
*/

// RSC_Service Service UUID
CONST uint8_t RSC_ServiceUUID[ATT_BT_UUID_SIZE] =
{
  LO_UINT16(RSC_SERVICE_SERV_UUID), HI_UINT16(RSC_SERVICE_SERV_UUID)
};

// RSC_Msr UUID
CONST uint8_t RSC_Service_RSC_MsrUUID[ATT_BT_UUID_SIZE] =
{
  LO_UINT16(RSC_SERVICE_RSC_MSR_UUID), HI_UINT16(RSC_SERVICE_RSC_MSR_UUID)
};

// RSC_Ftr UUID
CONST uint8_t RSC_Service_RSC_FtrUUID[ATT_BT_UUID_SIZE] =
{
  LO_UINT16(RSC_SERVICE_RSC_FTR_UUID), HI_UINT16(RSC_SERVICE_RSC_FTR_UUID)
};

/*********************************************************************
 * LOCAL VARIABLES
 */

static RSC_ServiceCBs_t *pAppCBs = NULL;

/*********************************************************************
* Profile Attributes - variables
*/

// Service declaration
static CONST gattAttrType_t RSC_ServiceDecl = { ATT_BT_UUID_SIZE, RSC_ServiceUUID };

// Characteristic "RSC_Msr" Properties (for declaration)
static uint8_t RSC_Service_RSC_MsrProps = GATT_PROP_NOTIFY;  //    GATT_PROP_READ | GATT_PROP_NOTIFY;

// Characteristic "RSC_Msr" Value variable
static uint8_t RSC_Service_RSC_MsrVal[RSC_SERVICE_RSC_MSR_LEN] = {0};

// Characteristic "RSC_Msr" CCCD
static gattCharCfg_t *RSC_Service_RSC_MsrConfig;

// Characteristic "RSC_Msr" User Description
//static uint8 RSC_Service_RSC_MsrUserDesc[4] = "User Description";

// Characteristic "RSC_Ftr" Properties (for declaration)
static uint8_t  RSC_Service_RSC_FtrProps = GATT_PROP_READ;

// Characteristic "RSC_Ftr" Value variable
static uint8_t  RSC_Service_RSC_FtrVal[RSC_SERVICE_RSC_FTR_LEN] = {0};

/*********************************************************************
* Profile Attributes - Table
*/

static gattAttribute_t RSC_ServiceAttrTbl[] =
{
  // RSC_Service Service Declaration
  {
    { ATT_BT_UUID_SIZE, primaryServiceUUID },
    GATT_PERMIT_READ,
    0,
    (uint8_t *)&RSC_ServiceDecl
  },
    // RSC_Msr Characteristic Declaration
    {
      { ATT_BT_UUID_SIZE, characterUUID },
      GATT_PERMIT_READ,
      0,
      &RSC_Service_RSC_MsrProps
    },
      // RSC_Msr Characteristic Value
      {
        { ATT_BT_UUID_SIZE, RSC_Service_RSC_MsrUUID },
        GATT_PERMIT_READ, // 0,
        0,
        RSC_Service_RSC_MsrVal
      },
      // RSC_Msr Client Characteristic Configuration Declaration
      {
        { ATT_BT_UUID_SIZE, clientCharCfgUUID },
        GATT_PERMIT_READ | GATT_PERMIT_WRITE,
        0,
        (uint8 *)&RSC_Service_RSC_MsrConfig
      },
    // RSC_Ftr Characteristic Declaration
    {
      { ATT_BT_UUID_SIZE, characterUUID },
      GATT_PERMIT_READ,
      0,
      &RSC_Service_RSC_FtrProps
    },
      // RSC_Ftr Characteristic Value
      {
        { ATT_BT_UUID_SIZE, RSC_Service_RSC_FtrUUID },
        GATT_PERMIT_READ,
        0,
        RSC_Service_RSC_FtrVal
      },
};

/*********************************************************************
 * LOCAL FUNCTIONS
 */
static bStatus_t RSC_Service_ReadAttrCB( uint16_t connHandle, gattAttribute_t *pAttr,
                                           uint8_t *pValue, uint16_t *pLen, uint16_t offset,
                                           uint16_t maxLen, uint8_t method );
static bStatus_t RSC_Service_WriteAttrCB( uint16_t connHandle, gattAttribute_t *pAttr,
                                            uint8_t *pValue, uint16_t len, uint16_t offset,
                                            uint8_t method );

/*********************************************************************
 * PROFILE CALLBACKS
 */
// RSCP Service Callbacks
CONST gattServiceCBs_t RSC_ServiceCBs =
{
  RSC_Service_ReadAttrCB,  // Read callback function pointer
  RSC_Service_WriteAttrCB, // Write callback function pointer
  NULL                       // Authorization callback function pointer
};

/*********************************************************************
* PUBLIC FUNCTIONS
*/

/*
 * RSC_Service_AddService- Initializes the RSC_Service service by registering
 *          GATT attributes with the GATT server.
 *
 */
extern bStatus_t RSC_Service_AddService( uint8_t rspTaskId )
{
  uint8_t status;

  // Allocate Client Characteristic Configuration table
  RSC_Service_RSC_MsrConfig = (gattCharCfg_t *)ICall_malloc( sizeof(gattCharCfg_t) * linkDBNumConns );
  if ( RSC_Service_RSC_MsrConfig == NULL )
  {
    return ( bleMemAllocError );
  }

  // Initialize Client Characteristic Configuration attributes
  GATTServApp_InitCharCfg( LINKDB_CONNHANDLE_INVALID, RSC_Service_RSC_MsrConfig );
  // Register GATT attribute list and CBs with GATT Server App
  status = GATTServApp_RegisterService( RSC_ServiceAttrTbl,
                                        GATT_NUM_ATTRS( RSC_ServiceAttrTbl ),
                                        GATT_MAX_ENCRYPT_KEY_SIZE,
                                        &RSC_ServiceCBs );

  return ( status );
}

/*
 * RSC_Service_RegisterAppCBs - Registers the application callback function.
 *                    Only call this function once.
 *
 *    appCallbacks - pointer to application callbacks.
 */
bStatus_t RSC_Service_RegisterAppCBs( RSC_ServiceCBs_t *appCallbacks )
{
  if ( appCallbacks )
  {
    pAppCBs = appCallbacks;

    return ( SUCCESS );
  }
  else
  {
    return ( bleAlreadyInRequestedMode );
  }
}

/*
 * RSC_Service_SetParameter - Set a RSC_Service parameter.
 *
 *    param - Profile parameter ID
 *    len - length of data to right
 *    value - pointer to data to write.  This is dependent on
 *          the parameter ID and WILL be cast to the appropriate
 *          data type (example: data type of uint16 will be cast to
 *          uint16 pointer).
 */
bStatus_t RSC_Service_SetParameter( uint8_t param, uint16_t len, void *value )
{
  bStatus_t ret = SUCCESS;
  switch ( param )
  {
    case RSC_SERVICE_RSC_MSR_ID:
      if ( len == RSC_SERVICE_RSC_MSR_LEN )
      {
        memcpy(RSC_Service_RSC_MsrVal, value, len);

        // Try to send notification.
        GATTServApp_ProcessCharCfg( RSC_Service_RSC_MsrConfig, (uint8_t *)&RSC_Service_RSC_MsrVal, FALSE,
                                    RSC_ServiceAttrTbl, GATT_NUM_ATTRS( RSC_ServiceAttrTbl ),
                                    INVALID_TASK_ID,  RSC_Service_ReadAttrCB);
      }
      else
      {
        ret = bleInvalidRange;
      }
      break;

    case RSC_SERVICE_RSC_FTR_ID:
      if ( len == RSC_SERVICE_RSC_FTR_LEN )
      {
        memcpy(RSC_Service_RSC_FtrVal, value, len);
      }
      else
      {
        ret = bleInvalidRange;
      }
      break;

    default:
      ret = INVALIDPARAMETER;
      break;
  }
  return ret;
}


/*
 * RSC_Service_GetParameter - Get a RSC_Service parameter.
 *
 *    param - Profile parameter ID
 *    value - pointer to data to write.  This is dependent on
 *          the parameter ID and WILL be cast to the appropriate
 *          data type (example: data type of uint16 will be cast to
 *          uint16 pointer).
 */
bStatus_t RSC_Service_GetParameter( uint8_t param, uint16_t *len, void *value )
{
  bStatus_t ret = SUCCESS;
  switch ( param )
  {
    default:
      ret = INVALIDPARAMETER;
      break;
  }
  return ret;
}


/*********************************************************************
 * @fn          RSC_Service_ReadAttrCB
 *
 * @brief       Read an attribute.
 *
 * @param       connHandle - connection message was received on
 * @param       pAttr - pointer to attribute
 * @param       pValue - pointer to data to be read
 * @param       pLen - length of data to be read
 * @param       offset - offset of the first octet to be read
 * @param       maxLen - maximum length of data to be read
 * @param       method - type of read message
 *
 * @return      SUCCESS, blePending or Failure
 */
static bStatus_t RSC_Service_ReadAttrCB( uint16_t connHandle, gattAttribute_t *pAttr,
                                       uint8_t *pValue, uint16_t *pLen, uint16_t offset,
                                       uint16_t maxLen, uint8_t method )
{
  bStatus_t status = SUCCESS;

  // See if request is regarding the RSC_Msr Characteristic Value
if ( ! memcmp(pAttr->type.uuid, RSC_Service_RSC_MsrUUID, pAttr->type.len) )
  {
    if ( offset > RSC_SERVICE_RSC_MSR_LEN )  // Prevent malicious ATT ReadBlob offsets.
    {
      status = ATT_ERR_INVALID_OFFSET;
    }
    else
    {
      *pLen = MIN(maxLen, RSC_SERVICE_RSC_MSR_LEN - offset);  // Transmit as much as possible
      memcpy(pValue, pAttr->pValue + offset, *pLen);
    }
  }
else if ( ! memcmp(pAttr->type.uuid, RSC_Service_RSC_FtrUUID, pAttr->type.len) )
  {
    if ( offset > RSC_SERVICE_RSC_FTR_LEN )  // Prevent malicious ATT ReadBlob offsets.
    {
      status = ATT_ERR_INVALID_OFFSET;
    }
    else
    {
      *pLen = MIN(maxLen, RSC_SERVICE_RSC_FTR_LEN - offset);  // Transmit as much as possible
      memcpy(pValue, pAttr->pValue + offset, *pLen);
    }
  }
  else
  {
    // If we get here, that means you've forgotten to add an if clause for a
    // characteristic value attribute in the attribute table that has READ permissions.
    *pLen = 0;
    status = ATT_ERR_ATTR_NOT_FOUND;
  }


  return status;
}


/*********************************************************************
 * @fn      RSC_Service_WriteAttrCB
 *
 * @brief   Validate attribute data prior to a write operation
 *
 * @param   connHandle - connection message was received on
 * @param   pAttr - pointer to attribute
 * @param   pValue - pointer to data to be written
 * @param   len - length of data
 * @param   offset - offset of the first octet to be written
 * @param   method - type of write message
 *
 * @return  SUCCESS, blePending or Failure
 */
static bStatus_t RSC_Service_WriteAttrCB( uint16_t connHandle, gattAttribute_t *pAttr,
                                        uint8_t *pValue, uint16_t len, uint16_t offset,
                                        uint8_t method )
{
  bStatus_t status  = SUCCESS;
  uint8_t   paramID = 0xFF;

  // See if request is regarding a Client Characterisic Configuration
  if ( ! memcmp(pAttr->type.uuid, clientCharCfgUUID, pAttr->type.len) )
  {
    // Allow only notifications.
    status = GATTServApp_ProcessCCCWriteReq( connHandle, pAttr, pValue, len,
                                             offset, GATT_CLIENT_CFG_NOTIFY);
  }
  else
  {
    // If we get here, that means you've forgotten to add an if clause for a
    // characteristic value attribute in the attribute table that has WRITE permissions.
    status = ATT_ERR_ATTR_NOT_FOUND;
  }

  // Let the application know something changed (if it did) by using the
  // callback it registered earlier (if it did).
  if (paramID != 0xFF)
    if ( pAppCBs && pAppCBs->pfnChangeCb )
      pAppCBs->pfnChangeCb(connHandle, paramID, len, pValue); // Call app function from stack task context.

  return status;
}
